<?php
class togmine_pull_time_entries extends togmine_synchronizer_base {
  /**
   * Initialise status variables.
   */
  public $success = FALSE;
  public $warning = FALSE;

  /**
   * Track various statistics about new and modified records.
   */
  public $new = 0;
  public $updated = 0;
  public $deleted = 0;
  public $ignored = 0;
  public $errors = 0;
  public $users = 0;

  /**
   * Prepare to execute the command.
   */
  function start() {
    // Time the processing.
    timer_start('togmine_pull_time_entries');
    $this->started = timer_read('togmine_pull_time_entries');
    $this->timer = array($this->started => 'started');
  }

  /**
   * Execute the command for one user account.
   */
  function execute() {
    $this->users++;

    // Get the time entries for this account.
    if ($entries = toggl_rest_api_time_entries($this->get_period())) {
      foreach ($entries as $item) {
        $entry = $this->process_data($item);
        if (!$entry) {
          // Do not count autocomplete suggestions in the "ignored" list.
          if ($entry === FALSE) {
            $this->ignored++;
          }
        }
        else {
          // Look for issue_mentions with no redmine issues, and alert the user.
          foreach (array_keys($entry->issue_mentions) as $id) {
            if (!isset($entry->redmine_issue_ids[$id])) {
              $entry->messages[] = t('Issue #!id does not exist in Redmine.', array('!id' => $id));
              // @todo Postponed: Work out which Redmine project maps to this
              // Toggl Project and save the time entry anyway (without an issue_id).
            }
          }

          if (empty($entry->redmine_time_entry_ids) && empty($entry->redmine_issue_ids)) {
            $this->ignored++;
          }
          else {
            $this->synchronize_time_entry($entry);
            // Keep track of the Toggl time entry IDs to handle orphans later.
            $this->synchronized_ids[] = $entry->toggl_time_entry_id;

            // For time entries that have a different number of redmine issue
            // references than they did at last sync, redmine time log entries are
            // deleted and the whole time entry is treated as a new time entry.
            // Count these as updates, not new time entries.
            if ($entry->deleted || $entry->update_mode) {
              $this->updated++;
            }
            else {
              $this->new++;
            }

            // Also count time entries that got deleted and reinserted.
            if ($entry->deleted) {
              $this->deleted++;
            }
          }

          if (count($entry->messages)) {
            $title = t('Warnings about "%description" (Toggl time entry !id):', array('!id' => $entry->toggl_time_entry_id, '%description' => $entry->description));
            $this->messages[] = $this->render_message_group($entry->messages, $title);
          }
        }
      }
    }

    if (count($result->messages)) {
      $this->messages[] = $this->render_message_group($result->messages, "\n## " . t('Messages for %name (%email):', array('%name' => $this->active_account->name, '%email' => $this->active_account->mail)) . "\n");
      $this->warning = TRUE;
    }

    // Log the time.
    $time = timer_read('togmine_pull_time_entries');
    $this->timer[$time] = $this->active_account->uid;
  }

  /**
   * Tidy up after the command has finished running and render statistics.
   */
  function end() {
    $this->handle_orphans();

    // Log the time pull_time_entries begins.
    $this->finished = time();

    // Stop the timer.
    $this->duration = timer_read('togmine_pull_time_entries');
    $this->timer[$this->duration] = 'finished';

    // Report on the status of the sync.
    // Set up args for messages.
    $args = array(
      '!new' => $this->new,
      '!updated' => $this->updated,
      '!deleted' => $this->deleted,
      '!ignored' => $this->ignored,
      '!errors' => $this->errors,
      '!orphaned_redmine_time_entries' => count($this->orphaned_redmine_time_entries),
      '!users' => $this->users,
      '!date' => $this->get_period('start'),
      '!time' => round($this->duration / 1000),
    );

    if (count($this->orphaned_redmine_time_entries)) {
      $orphans_messages = $this->orphans_messages();
      $title = "\n## " . t('!orphaned_redmine_time_entries Redmine time entries are orphaned:', $args);
      $this->messages[] = $this->render_message_group($orphans_messages, $title);
      $this->warning = TRUE;
    }

    if (!empty($this->synchronized_ids)) {
      $this->success = TRUE;
    }

    if ($this->ignored) {
      // It is expected that there will always be some time entries that are
      // ignored.  Internal-billable time log entries almost never have ticket
      // numbers associated.
      // @todo Postponed: Alert users about Toggl time entries in Toggl projects
      // that are flagged billable, if the time entry description is missing ticket
      // references.
      $this->messages[] = t('!ignored time entries were ignored.', $args);
    }

    if ($this->errors) {
      $this->messages[] = t('!errors data-write errors were encountered.', $args);
      $this->success = FALSE;
    }

    $title = t('!new new & !updated existing time entries since !date from !users users, were synchronized in !time seconds.', $args);
    $this->message = empty($this->messages) ? $title : $this->render_message_group(array_reverse($this->messages), $title);
  }

  /**
   * Get dates relating to the backlog period; start_date, end_date or both.
   *
   * @param $start_or_end String
   *    Set to either 'start' or 'end' to return only the respctive date.
   *    Omit to get both in an Object.
   * @return Mixed
   *    String Date if $start_or_end is set and is valid.
   *    Object with two properties for start_date and end_date respectively.
   */
  function get_period($start_or_end = NULL) {
    static $period;

    // Store the period in a static, so that the start and end period remain
    // constant for the life of this synchronization.
    if (!$period) {
      $now = time();
      $period = toggl_date_parameters($now - togmine_var('backlog_period'), $now);
    }

    if ($start_or_end) {
      return $period["{$start_or_end}_date"];
    }

    return $period;
  }

  /**
   * Synchronize one pre-processed Toggl Time entry to Redmine & update the map.
   */
  function synchronize_time_entry(&$entry) {
    // How many Redmine time entries do we know about for this Toggl.com time
    // entry?
    $known = count($entry->redmine_time_entry_ids);
    // How many of the issues that were mentioned in the Toggl description are
    // valid?
    $valid = count($entry->redmine_issue_ids);
    if ($known === $valid) {
      // This is a normal update of the columns only.  No rows need to to be added
      // or removed.
      // Reset the counter for consecutive repeats of this occuring, instead of
      // simply DELETEing the row for better logging and debugging.
      // If a count has never been set for this Toggl.com time entry this will not change any rows.
      db_query('UPDATE {togmine_time_entry_delete_counts} SET count = 0 WHERE tid = %d', $entry->toggl_time_entry_id);
    }
    // Filter out brand new time entries.
    else if ($known > 0) {
      // The number of issues mentioned in the Toggl time entry description is
      // different to the number of Redmine time entries TogMine has for this Toggl
      // time entry.  This can occur if any of the following have occured since
      // TogMine last synchronized this time entry;
      //  * Some of the issues did not exist in Redmine but do now.
      //  * Issues have been added or removed in the Toggl time entry description.
      //  * The date of the Toggl time entry has changed.
      // In all of these scenarios, delete all the Redmine time entries and insert
      // new ones.

      $this->entry_delete($entry);
      $entry->deleted = TRUE;
      $entry->redmine_time_entry_ids = array();

      // Check if this happened last time TogMine synchronized.
      $count = db_result(db_query('SELECT count FROM {togmine_time_entry_delete_counts} WHERE tid = %d', $entry->toggl_time_entry_id));
      if ($count === FALSE) {
        // It is the first time this happened.  Just note it for next time.
        db_query('INSERT INTO {togmine_time_entry_delete_counts} VALUES (%d, 1)', $entry->toggl_time_entry_id);
      }
      else {
        // This happened last time TogMine synchronized.  Increment the counter and log it.
        db_query('UPDATE {togmine_time_entry_delete_counts} SET count = %d WHERE tid = %d', ++$count, $entry->toggl_time_entry_id);
        if ($count > 1) {
          $entry->messages[] = t('!count consecutive times, the number of valid issues referenced in the Toggl time entry (!valid) differed from the number that TogMine synchronized last time (!known).', array('!count' => $count, '!valid' => $valid, '!known' => $known));
        }
      }
    }

    $entry->update_mode = (Boolean) count($entry->redmine_time_entry_ids);

    // entry_save() will update all the time entries anyway, regardless of whether the issue ID numbers or the order of the issue IDs has changed, or whether the time entry was updated in Redmine.
    $this->entry_save($entry, $entry->update_mode);
  }

  /**
   * Looks for Redmine time entries that have no entry on Toggl.com.
   *
   * This occurs when;
   *  - The time entry belongs to a user that no longer has a valid Toggl API token
   *    or Redmine user ID.
   *  - The Toggl time entry may have been deleted in Toggl.
   * @todo Are there any other cases that cause time entries to be orphaned?
   */
  function handle_orphans() {
    $this->orphaned_redmine_time_entries = array();

    // Time entries belonging to blocked users should not be reported as orphans.
    $sql = "SELECT t.redmine_time_entry_id FROM {users} u JOIN {togmine_time_entries} t ON t.uid = u.uid WHERE u.status = 1 AND t.spent_on > '%s'";
    $args = array($this->get_period('start'));

    if (!empty($this->synchronized_ids)) {
      $sql .= ' AND toggl_time_entry_id NOT IN (%s)';
      $args[] = implode(',', $this->synchronized_ids);
    }
    $result = db_query($sql, $args);

    while ($row = db_fetch_object($result)) {
      $id = $row->redmine_time_entry_id;
      if ($entry = redmine_rest_api_time_entry_load($id)) {
        $this->orphaned_redmine_time_entries[$id] = $entry;
      }
      else {
        // The time entry is missing in both Toggl, and Redmine.  Remove it from
        // Togmine's map.  Do not permanently delete it though, since this case may
        // occur if both web services are temporarily and legitimately unavailable.
        $this->map_archive($id);
      }
    }
  }

  /**
   * Render messages about orphans.
   */
  function orphans_messages() {
    $this->messages = array();
    // Re-use the same translatable string but with empty arguments to work out how
    // long the prefix is on redmine time entry comments.
    $prefix_length = strlen(t('Toggl: !comments', array('!comments' => '')));
    foreach ($this->orphaned_redmine_time_entries as $entry) {
      $args = array(
        '!time_spent' => round($entry->hours, 2),
        '%date' => $entry->spent_on,
        '!name' => $entry->user->name,
        // Exclude "Toggl: " from the beginning of the description.
        '%toggl_description' => substr($entry->comments, $prefix_length),
        '!project' => $entry->project->name,
        '!issue_id' => $entry->issue->id,
        '%issue_title' => redmine_rest_api_issue_load($entry->issue->id)->subject,
      );
      $this->messages[] = t('!time_spent hours on %date by !name on "%toggl_description" for !project (#!issue_id %issue_title)', $args);
    }
    return $messages;
  }



  /***************************************************************************
   * TogMine API functions.
   ***************************************************************************/

  /**
   * Processes raw Toggl.com time entries to make them more useful for TogMine.
   */
  function process_data($data) {
    // Filter out time entries that are from the wrong workspace.
    if (!toggl_workspace_id_filter($data)) {
      return FALSE;
    }

    // Filter out time entries where the timer is still ticking.
    if ($data->duration < 0) {
      return FALSE;
    }

    // Filter out autocomplete suggestions, which are not really time entries.
    if (db_result(db_query('SELECT redmine_issue_id FROM {togmine_issues} WHERE toggl_time_entry_id = %d', $data->id))) {
      return;
    }

    // Create a new TogMine entry.
    $entry = new stdClass();
    $entry->toggl_time_entry_id = $data->id;
    $entry->time = $data->duration;
    $entry->description = $data->description;
    $entry->billable = $data->billable;
    $entry->date = $data->start;
    $entry->redmine_time_entry_ids = $this->redmine_entry_ids($data->id);
    // Get all the references to issue numbers.
    $entry->issue_mentions = $this->grep_issues($data->description);
    // Filter out the issues that actually *exist* in Redmine.
    $entry->redmine_issue_ids = $this->redmine_issues_exist(array_keys($entry->issue_mentions));
    $entry->messages = array();
    $entry->deleted = FALSE;
    return $entry;
  }

  /**
   * Saves a Toggl entry's time entries to Redmine and the TogMine map.
   *
   * @param $data Object
   *    The time entry to save.  If $data->update_mode is TRUE,
   *    $data->redmine_time_entry_ids MUST have the same number of elements as
   *    $data->redmine_issue_ids, or the save will fail.
   * @return Integer
   *    The number of database errors.  0 on success.
   */
  function entry_save(&$data) {
    // Set the variables that will apply to all Redmine time entries belonging to
    // this Toggl entry.
    $comments = t('Toggl: !comments', array('!comments' => $data->description));

    // @todo Postponed: Use an enumerations API to get a valid default activity ID.
    // @see http://www.redmine.org/issues/7819
    $activity_id = togmine_var('redmine_default_activity');

    // How many times was a ticket mentioned (unique or not)?  E.g. "Worked on #1
    // #2 and #2" counts as 3 mentions.
    $total_mentions = array_sum($data->issue_mentions);
    // What is the rounded time, in seconds?
    $rounded = $this->round($data->time);
    // How many hours does each mention get, as a fraction?
    $hours_per_mention = ($total_mentions < 1) ? 0 : $rounded / $total_mentions / 60 / 60;
    // If 2 tickets are mentioned but one is not in Redmine, the other ticket still
    // only gets it's share of time logged to it.

    // Assert that count($data->redmine_time_entry_ids) == count($data->issue_mentions)
    if ($data->update_mode && count($data->redmine_time_entry_ids) !== count($data->redmine_issue_ids)) {
      // This indicates a severe bug in TogMine.
      rve($data, "For Toggl.com time entry ID $data->id there are $num_time_entries time entries and $num_issues.");
      return FALSE;
    }

    // An index counter to access $data->redmine_time_entry_ids.
    $i = 0;
    foreach ($data->redmine_issue_ids as $issue_id) {
      // Build and save a redmine_rest_api_time_entry_save()-compatible $entry object.
      // @see http://drupal.org/node/942024#comment-3574866
      $entry = new stdClass();
      $old_id = FALSE;

      if ($data->update_mode) {
        $redmine_time_entry_id = $data->redmine_time_entry_ids[$i++];
        // Verify the Redmine time-entry exists.  A Redmine user may have deleted it.
        if ((Boolean) redmine_rest_api_time_entry_load($redmine_time_entry_id)) {
          $entry->id = $redmine_time_entry_id;
        }
        else {
          // Someone deleted the time entry that TogMine created on last sync.
          // @todo Minor: Look in Redmine's logs to find who deleted the time
          // entry and scold them.
          $data->messages[] = t('Time entry on #!issue_id for "%description" was deleted in Redmine.', array('!issue_id' => $issue_id, '%description' => $data->description));
          $old_id = $redmine_time_entry_id;
        }
      }

      $entry->issue_id = $issue_id;
      $entry->hours = $data->issue_mentions[$issue_id] * $hours_per_mention;
      $entry->comments = $comments;
      $entry->activity_id = $activity_id;
      $entry->spent_on = $data->date;

      // Save the time log entry and log any errors.
      if (!$response = redmine_rest_api_time_entry_save($entry)) {
        // Do not update the map if there is no valid response.
        $this->errors++;
        rest_api_query_var_export(compact('response', 'entry'), 'No response from redmine_rest_api_time_entry_save()');
      }
      else {
        // Build and save the map to the togmine table.
        $map = new stdClass();
        $map->redmine_time_entry_id = empty($entry->id) ? $response->id : $entry->id;
        $map->toggl_time_entry_id = $data->toggl_time_entry_id;
        $map->uid = $this->active_account->uid;
        $map->spent_on = date(TOGGL_DATE_FORMAT, strtotime($entry->spent_on));

        // Save this map to the database.
        $this->map_save($map, $data->update_mode, $old_id);
      }
    }
  }

  /**
   * Delete a time entry from Redmine and it's corresponding map.
   */
  function entry_delete($entry) {
    $from = 'FROM {togmine_time_entries} WHERE toggl_time_entry_id = %d';
    $result = db_query("SELECT redmine_time_entry_id $from", $entry->toggl_time_entry_id);
    while ($redmine_time_entry_id = db_result($result)) {
      redmine_rest_api_time_entry_delete($redmine_time_entry_id);
    }
    db_query("DELETE $from", $entry->toggl_time_entry_id);
  }

  /**
   * Save a map to the database.
   *
   * @param $map stdClass
   *    Must have properties for each column of the {togmine_time_entries} table:
   *    * redmine_time_entry_id
   *    * toggl_time_entry_id
   *    * uid
   *    * spent_on
   * @param $update Boolean
   *    Whether this is an update or not
   * @param $old_id Integer
   *    What the old Redmine time entry ID was if it has changed.
   * @ret
   */
  function map_save(&$map, $update, $old_id = FALSE) {
    // Save the map and log any errors.
    if (!$update) {
      drupal_write_record('togmine_time_entries', $map) !== SAVED_NEW ? $this->errors++ : NULL;
    }
    else if (!$old_id) {
      drupal_write_record('togmine_time_entries', $map, 'redmine_time_entry_id') !== SAVED_UPDATED ? $this->errors++ : NULL;
    }
    else {
      // Update manually, since drupal_write_record() can not update the primary key.
      // drupal_write_record() makes a fair assumption that the primary key for
      // the table is a serial column (or used in an equivalent manner).  However
      // in {togmine_time_entries} the primary key is a foreign key to
      // redmine.time_entries.id.
      $sql = 'UPDATE {togmine_time_entries} SET redmine_time_entry_id = %d, toggl_time_entry_id = %d, spent_on = %s WHERE redmine_time_entry_id = %d';
      $args = array($map->redmine_time_entry_id, $map->toggl_time_entry_id, $map->spent_on, $old_id);
      db_query($sql, $args) === FALSE ? $this->errors++ : NULL;
    }
  }

  /**
   * Archive a time entry map.  Never delete them.
   */
  function map_archive($redmine_time_entry_id) {
    $from = 'FROM {togmine_time_entries} WHERE redmine_time_entry_id = %d';
    // Copy the record to {togmine_time_entries_archive} before deleting it.
    db_query("INSERT INTO {togmine_time_entries_archive} SELECT *, %d AS timestamp $from", time(), $redmine_time_entry_id);
    db_query("DELETE $from", $redmine_time_entry_id);
  }

  /***************************************************************************
   * TogMine helper functions.
   */

  /**
   * Gets possible Redmine Issue IDs from the Toggl entry's description.
   */
  function grep_issues($string) {
    $issues = array();

    // Make sure we don't parse strings like "s#123" and "#123-foo".
    foreach (preg_split('/\s/', $string) as $word) {
      // Allow for punctuation.  A description like "#2, #3 and not #7f -- finally
      // -- #4!" should match issues 2, 3 and 4 but not 7.
      if (preg_match('/^[\(\[]?#([0-9]+)[.,:;\)\]!\?]{0,2}$/', $word, $matches)) {
        $id = (Integer) $matches[1];
        // A ticket may be mentioned multiple times in one description.
        $issues[$id]++;
      }
    }

    return $issues;
  }

  /**
   * Load the Project IDs that correspond to a number of issue IDs.
   */
  function redmine_issues_exist($issue_ids) {
    $issues = array();
    foreach ($issue_ids as $id) {
      if (redmine_rest_api_issue_head($id)) {
        $issues[$id] = $id;
      }
    }
    return $issues;
  }

  /**
   * Gets the Redmine Time Entry IDs for a given Toggl time entry ID.
   */
  function redmine_entry_ids($toggl_time_entry_id) {
    $redmine_ids = array();
    // SQL to get the Redmine time entry IDs for Toggl entry ID.
    $result = db_query('SELECT redmine_time_entry_id FROM {togmine_time_entries} WHERE toggl_time_entry_id = %d', $toggl_time_entry_id);
    while ($redmine_time_entry_id = db_result($result)) {
      $redmine_ids[] = $redmine_time_entry_id;
    }
    return $redmine_ids;
  }

  /**
   * Rounds a duration in seconds up to time blocks.
   *
   * Rounds up to the next time block, unless $seconds is less than X greater
   * than the previous time block.  The threshold (X) and the length of the time
   * blocks are configurable.
   *
   * @see variable_get('togmine_rounding_block')
   * @see variable_get('togmine_rounding_threshold')
   *
   * @param $seconds Integer
   *    The duration in seconds.
   * @return Integer
   *    The rounded duration in seconds.
   */
  static function round($seconds) {
    $block_len = (Float) togmine_var('rounding_block');
    $threshold = togmine_var('rounding_threshold') / $block_len;

    // Convert Toggl time in seconds to a number of time blocks, as a fraction.
    $blocks = $seconds / $block_len;

    // Round up to the nearest time block, an integer.
    $rounded = ceil($blocks);

    // If the amount that was rounded is less than the threshold, round down instead.
    if ((1 + $blocks - $rounded) < $threshold) {
      $rounded--;
    }

    // Convert number of time blocks back to seconds.
    return $rounded * $block_len;
  }

  /**
   * Formats and renders a group of messages into an easily readable list.
   */
  function render_message_group($messages, $title) {
    // Add some new-lines characters around the title so that it formats nicely on the command line too.  For example with drush_log() + filter_xss().
    // @see http://drupal.org/node/987852
    return theme('item_list', $messages, "$title\n", 'ol');
  }
}
