<?php
/**
 * TogMine synchronization command to push projects from Redmine to Toggl.
 */
class togmine_push_projects extends togmine_synchronizer_base {
  /**
   * Track the workspaces that have projects have been pushed too.
   *
   * Multiple accounts may use the same workspace.
   */
  public $synchronized_workspaces = array();

  /**
   * Override synchronize_all_accounts().
   *
   * If this Drupal website is configured to use a shared workspace, execute()
   * should only run once using the global API keys for the shared workspace, and
   * does not synchronize projects for each user.
   *
   * @see http://drupal.org/node/1346572
   */
  function synchronize_all_accounts() {
    if (variable_get('toggl_workspace_id', FALSE)) {
      // Users of this Drupal site uses a shared Toggl workspace.  Do not iterate
      // over accounts.
      $this->execute();
    }
    else {
      parent::synchronize_all_accounts();
    }
  }

  /**
   * Do the synchronization.
   */
  function execute() {
    // What workspace needs to be synchronized for this user?
    $workspace_id = $this->get_workspace_id();
    if (isset($this->synchronized_workspaces[$workspace_id])) {
      // This workspace was already synchronized in this synchronization.
      return;
    }

    // Track which workspaces have been synchronized.
    $this->synchronized_workspaces[$workspace_id] = $workspace_id;

    // Get all the Redmine projects for this user/workspace & iterate over them.
    if ($redmine_result = redmine_rest_api_projects()) {
      foreach ($redmine_result as $redmine) {
        // Look for the Toggl ID for this Redmine project if it exists.
        $sql = 'SELECT toggl_project_id FROM {togmine_projects} WHERE redmine_project_id = %d AND toggl_workspace_id = %d';
        $id = db_result(db_query($sql, $redmine->id, $workspace_id));

        // Create a new empty stdClass object if this is a new project.
        $project = $id ? toggl_rest_api_project_load($id) : new stdClass;

        // Update/overwrite the Toggl project name, even if it is set already.
        $project->name = $redmine->name;

        // Set the workspace if it is not already set.
        if (empty($project->workspace)) {
          $project->workspace_id = $this->get_workspace_id();
        }

        // Save the project to Toggl.
        if ($saved = toggl_rest_api_project_save($project)) {
          $this->synchronized_ids[] = $saved->id;
          // If this is a new project, save it's mapping to the database.
          if (!$id) {
            $sql = "INSERT INTO {togmine_projects} (redmine_project_id, toggl_project_id, toggl_workspace_id) VALUES (%d, %d, %d)";
            db_query($sql, $redmine->id, $saved->id, $workspace_id);
          }
        }
        else {
          rve(t('Redmine project !name could not be saved to Toggl.com due to an unknown API error.', array('!name' => $redmine->name)), 'TogMine Push Projects');
        }
      }
    }

    // Delete orphaned Toggl projects from the map.
    // The projects can not be deleted from Toggl.com through the API.
    if (!empty($this->synchronized_ids)) {
      $ids = implode(',', $this->synchronized_ids);
      $from = "FROM {togmine_projects} WHERE toggl_workspace_id = %d AND toggl_project_id NOT IN ($ids)";

      // Check if there are any to delete and notify the administrator.
      if (db_result(db_query("SELECT count(*) $from", $workspace_id))) {
        $this->message = t('One or more projects were deleted in Redmine.  TogMine can not delete or archive it\'s corresponding Toggl.com project via the API.  Consider archiving or deleting the Toggl project via the Toggl.com user interface: https://www.toggl.com/projects');
      }

      // Delete the map items.
      db_query("DELETE $from", $workspace_id);
    }
  }
}
