<?php
/**
 * A base class for TogMine synchronization commands.
 */
abstract class togmine_synchronizer_base {
  /**
   * The IDs of the items that were created or updated in the destination.
   */
  public $synchronized_ids = array();

  /**
   * Call this method to execute a synchronization.
   */
  public function synchronize($account = NULL) {
    $this->start();

    // Synchronize an individual user account?
    if ($account) {
      // Yes.
      $this->synchronize_account($account);
    }
    else {
      // No.
      $this->synchronize_all_accounts();
    }

    $this->end();
  }

  /**
   * Handle the case where all Drupal user accounts should be synchronized.
   */
  protected function synchronize_all_accounts() {
    foreach ($this->get_accounts() as $account) {
      $this->synchronize_account($account);
    }
  }

  /**
   * Synchronize a single Drupal user account.
   */
  protected function synchronize_account($account) {
    $this->set_active_account($account);
    $this->execute();
    $this->set_active_account();
  }

  /**
   * Sub-classes must implement execute() which executes for each account.
   */
  public abstract function execute();

  /**
   * Sub-classes may implement the start() and end() methods.
   *
   * They run before and after a synchronization.
   */
  public function start() {}
  public function end() {}

  /**
   * Get users with both Toggl API tokens and Redmine user IDs.
   *
   * @return Array
   *    Each item is a pseudo-user object with the following properties:
   *    * uid: The Drupal user ID.
   *    * toggl_api_token: The Toggle API token.
   *    * redmine_api_access_key: The Redmine API Acess key.
   *    * name: The Drupal user name.
   *    * mail: The email address of the suer, according to Drupal.
   */
  public static function get_accounts() {
    // Start building the array of arguments for the main query.
    $user_fields = array('u.uid', 'u.name', 'u.mail');
    $profile_fields = array('toggl_api_token', 'redmine_api_access_key');
    return get_multiple_user_access_keys($profile_fields, $user_fields);
  }

  /**
   * Sets the active account for an invocation of the execute() method.
   */
  public function set_active_account($account = NULL) {
    if ($account) {
      $this->active_account = $account;
      togmine_set_active_account($account);
    }
    else {
      $this->active_account = NULL;
      togmine_set_active_account();
    }
  }

  /**
   * Get a workspace object to use with project and issue objects.
   *
   * Defaults to $conf['toggl_workspace_id'] or current user's default workspace ID.
   */
  public static function get_workspace_id() {
    // Get the workspace ID variable.
    $id = variable_get('toggl_workspace_id', FALSE);

    // Use the user's default workspace ID if there is no workspace ID variable.
    if (!$id) {
      $id = get_toggl_user()->default_workspace_id;
    }

    return $id;
  }
}
